type segment_descriptor = {
    addr : int32 ;
    size : int32
}

type header = {
    (* Magic string : "PS-X EXE" *)
    magic : string ;

    (* Base of the text segment *)
    text_base : int32 ;

    (* Base of the data segment *)
    data_base : int32 ;

    (* First value of the PC "register" *)
    init_pc : int32 ;

    (* First value of the GP register *)
    init_gp : int32 ;

    (* Text, data, bss and stack address and size *)
    text : segment_descriptor ;
    data : segment_descriptor ;
    bss : segment_descriptor ;
    stack : segment_descriptor
}

type t = {
    header : header ;
    code : int32 array
}

module Make (FS : Filesystem.S) = struct
    let sector_size = 2048

    exception EOF
    let read s size =
        let ret, read = FS.read s size in
        if read <> size then
            raise EOF ;
        ret

    (* Read an int32 from the stream *)
    let read_int32 s =
        let i32_s = read s 4 in
        let bytes = Array.init 4 (fun i -> Char.code i32_s.[i]) in
        let lower = bytes.(0) + (bytes.(1) lsl 8) + (bytes.(2) lsl 16) in
        let upper = Int32.shift_left (Int32.of_int bytes.(3)) 24 in
        Int32.add (Int32.of_int lower) upper

    (* Read the whole header *)
    let read_header s =
        let read_segdescr () =
            let addr = read_int32 s in
            let size = read_int32 s in
            { addr = addr ; size = size }
        in
        let magic = read s 8 in
        let text_base = read_int32 s in
        let data_base = read_int32 s in
        let init_pc = read_int32 s in
        let init_gp = read_int32 s in
        let text = read_segdescr () in
        let data = read_segdescr () in
        let bss = read_segdescr () in
        let stack = read_segdescr () in {
            magic = magic ;
            text_base = text_base ;
            data_base = data_base ;
            init_pc = init_pc ;
            init_gp = init_gp ;
            text = text ;
            data = data ;
            bss = bss ;
            stack = stack
        }

    let parse stream =
        FS.seek stream Filesystem.Begin ;
        let hdr = read_header stream in
        FS.seek stream (Filesystem.Absolute sector_size) ;
        let code = ref [] in
        let eof = ref false in
        while not !eof do
            try
                code := (read_int32 stream) :: !code
            with EOF ->
                eof := true
        done ;
        { header = hdr ; code = Array.of_list (List.rev !code) }
end
