(*
 * Implementation of the R3000 instruction decoder.
 *
 * Basically, it's just int32 -> instr.
 *)

module Reg = R3000_reg

let bits_get_reg i32 x y =
    let b = Bits.get i32 x y in
    Reg.of_int b

(*
 * Instruction takes <= two registers and an immediate value.
 *)
module I = struct
    type args = {
        op : int ;
        rs : Reg.t ;
        rt : Reg.t ;
        imm : int
    }

    let decode i32 = {
        op = Bits.get i32 31 26 ;
        rs = bits_get_reg i32 25 21 ;
        rt = bits_get_reg i32 20 16 ;
        imm = Bits.get i32 15 0
    }

    let to_string name a =
        Printf.sprintf "%s\t\t%s, %s, 0x%08X" name (Reg.to_string a.rs)
            (Reg.to_string a.rt) a.imm
end

(*
 * Instruction is a unconditional jump.
 *)
module J = struct
    type args = {
        op : int ;
        target : int
    }

    let decode i32 = {
        op = Bits.get i32 31 26 ;
        target = Bits.get i32 25 0
    }

    let to_string name a =
        Printf.sprintf "%s\t\t0x%08X" name a.target
end

(*
 * Instruction operates on registers.
 *)
module R = struct
    type args = {
        op : int ;
        rs : Reg.t ;
        rt : Reg.t ;
        rd : Reg.t ;
        shift : int ;
        funct : int
    }

    let decode i32 = {
        op = Bits.get i32 31 26 ;
        rs = bits_get_reg i32 25 21 ;
        rt = bits_get_reg i32 20 16 ;
        rd = bits_get_reg i32 15 11 ;
        shift = Bits.get i32 10 6 ;
        funct = Bits.get i32 5 0
    }

    let to_string name a =
        Printf.sprintf "%s\t\t%s, %s, %s, 0x%08X"
            name (Reg.to_string a.rs) (Reg.to_string a.rt)
            (Reg.to_string a.rd) a.shift
end

(*
 * Instruction is a SYSCALL.
 *)
module S = struct
    type args = {
        op : int ;
        code : int ;
        funct : int
    }

    let decode i32 = {
        op = Bits.get i32 31 26 ;
        code = Bits.get i32 25 6 ;
        funct = Bits.get i32 5 0;
    }

    let to_string name a =
        Printf.sprintf "%s\t\t0x%08X" name a.code
end

(*
 * Instruction branches on some register condition.
 *)
module B = struct
    type args = {
        op : int ;
        rs : Reg.t ;
        typ : int ;
        imm : int
    }

    let decode i32 = {
        op = Bits.get i32 31 26 ;
        rs = bits_get_reg i32 25 21 ;
        typ = Bits.get i32 20 16 ;
        imm = Bits.get i32 15 0
    }

    let to_string name a =
        Printf.sprintf "%s\t\t%s, 0x%08X" name (Reg.to_string a.rs) a.imm
end

(*
 * Include the R3000 instructions from the inline macro file.
 *)
type instr =
#define R3000_INSTR(constr, m) \
|   constr of m.args
#include "instrs.inl"
#undef R3000_INSTR

(*
 * Makes a string of an R3000 instruction.
 *)

let to_string instr = match instr with
#define R3000_INSTR(constr, m) \
|   constr a -> m.to_string (String.lowercase #constr) a
#include "instrs.inl"
#undef R3000_INSTR

(*
 * Decode R3000 "special" instructions (code = 000000).
 *
 * Distinguishes the instructions according to the last six bits.
 *)
let decode_special i32 = match (Bits.get i32 5 0) with
|   0 -> Sll (R.decode i32)
|   2 -> Srl (R.decode i32)
|   3 -> Sra (R.decode i32)
|   4 -> Sllv (R.decode i32)
|   6 -> Srlv (R.decode i32)
|   7 -> Srav (R.decode i32)
|   8 -> Jr (R.decode i32)
|   9 -> Jalr (R.decode i32)
|   12 -> Syscall (S.decode i32)
|   13 -> Break (S.decode i32)
|   16 -> Mfhi (R.decode i32)
|   17 -> Mthi (R.decode i32)
|   18 -> Mflo (R.decode i32)
|   19 -> Mtlo (R.decode i32)
|   24 -> Mult (R.decode i32)
|   25 -> Multu (R.decode i32)
|   26 -> Div (R.decode i32)
|   27 -> Divu (R.decode i32)
|   32 -> Add (R.decode i32)
|   33 -> Addu (R.decode i32)
|   34 -> Sub (R.decode i32)
|   35 -> Subu (R.decode i32)
|   36 -> And (R.decode i32)
|   37 -> Or (R.decode i32)
|   38 -> Xor (R.decode i32)
|   39 -> Nor (R.decode i32)
|   42 -> Slt (R.decode i32)
|   43 -> Sltu (R.decode i32)
|   _ ->
        let s = Printf.sprintf "R3000_decoder.decode_special (b0-b5 = %d)"
            (Bits.get i32 5 0) in
        invalid_arg s

(*
 * Decode a branch instruction : code = 000001.
 *)
let decode_branch i32 = match (Bits.get i32 20 16) with
|   0 -> Bltz (B.decode i32)
|   1 -> Bgez (B.decode i32)
|   16 -> Bltzal (B.decode i32)
|   17 -> Bgezal (B.decode i32)
|   _ ->
        let s = Printf.sprintf "R3000_decoder.decode_branch (b20-b16 = %d)"
            (Bits.get i32 20 16) in
        invalid_arg s

(*
 * Main function of the module : decodes an arbitrary R3000 instruction
 * from an int32 to a R3000_decoder.instr.
 *)
let decode i32 = match (Bits.get i32 31 26) with
|   0 -> decode_special i32
|   1 -> decode_branch i32
|   2 -> J (J.decode i32)
|   3 -> Jal (J.decode i32)
|   4 -> Beq (I.decode i32)
|   5 -> Bne (I.decode i32)
|   6 -> Blez (B.decode i32)
|   7 -> Bgtz (B.decode i32)
|   8 -> Addi (I.decode i32)
|   9 -> Addiu (I.decode i32)
|   10 -> Slti (I.decode i32)
|   11 -> Sltiu (I.decode i32)
|   12 -> Andi (I.decode i32)
|   13 -> Ori (I.decode i32)
|   14 -> Xori (I.decode i32)
|   15 -> Lui (I.decode i32)
|   32 -> Lb (I.decode i32)
|   33 -> Lh (I.decode i32)
|   34 -> Lwl (I.decode i32)
|   35 -> Lw (I.decode i32)
|   36 -> Lbu (I.decode i32)
|   37 -> Lhu (I.decode i32)
|   38 -> Lwr (I.decode i32)
|   40 -> Sb (I.decode i32)
|   41 -> Sh (I.decode i32)
|   42 -> Swl (I.decode i32)
|   43 -> Sw (I.decode i32)
|   46 -> Swr (I.decode i32)
|   _ ->
        let s = Printf.sprintf "R3000_decoder.decode (b31-b26 = %d)"
            (Bits.get i32 31 26) in
        invalid_arg s
