/*
 * CPU/GPU plugin communication.
 */

#include "../config.h"
#include "gpu.h"
#include "memory.h"
#include "msg.h"

#include <dlfcn.h>
#include <stdlib.h>

/* Handle to the shared object */
static void* so;

static unsigned long display;

int (*psx_gpu_init)(void);
int (*psx_gpu_open)(unsigned long*, const char*, const char*);
int (*psx_gpu_close)(void);
int (*psx_gpu_shutdown)(void);
uint32_t (*psx_gpu_read_data)(void);
void (*psx_gpu_read_data_mem)(uint32_t*, int);
uint32_t (*psx_gpu_read_status)(void);
void (*psx_gpu_write_data)(uint32_t);
void (*psx_gpu_write_data_mem)(uint32_t*, int);
void (*psx_gpu_write_status)(uint32_t);
void (*psx_gpu_dma_chain)(uint32_t*, uint32_t);

void psx_gpu_load(void)
{
    so = dlopen(LIBDIR "/libgpu.so", RTLD_NOW);
    if (!so)
        psx_fail("No GPU plugin installed on this system !\n");

#define LOAD_(sym, sosym) \
    psx_gpu_##sym = dlsym(so, "GPU" #sosym); \
    if (!psx_gpu_##sym) \
        psx_fail("Unable to find symbol %s in the GPU plugin !\n", #sosym)

#define LOAD(sym) LOAD_(sym, sym)

    LOAD(init);
    LOAD(open);
    LOAD(close);
    LOAD(shutdown);
    LOAD_(read_data, readData);
    LOAD_(read_data_mem, readDataMem);
    LOAD_(read_status, readStatus);
    LOAD_(write_data, writeData);
    LOAD_(write_data_mem, writeDataMem);
    LOAD_(write_status, writeStatus);
    LOAD_(dma_chain, dmaChain);

    psx_gpu_init();
    psx_gpu_open(&display, "Meta-morphosed game", NULL);
}

void psx_gpu_unload(void)
{
    psx_gpu_close();
    psx_gpu_shutdown();

    dlclose(so);
}

unsigned long* psx_gpu_get_display(void)
{
    return &display;
}

#define DMA_GPU_TO_RAM 0x01000200
#define DMA_RAM_TO_GPU 0x01000201
#define DMA_CHAIN      0x01000401

void psx_gpu_do_dma(uint32_t madr, uint32_t bcr, uint32_t chcr)
{
    uint32_t size = (bcr >> 16) * (bcr & 0xFFFF);

    if (chcr == DMA_RAM_TO_GPU)
    {
        uint32_t lb = madr, ub = madr + size, i = 0;
        uint32_t* dmabuf;

        /* TODO: We could avoid the copy here */
        dmabuf = malloc(size * sizeof (uint32_t));
        while (lb < ub)
            dmabuf[i++] = psx_mem_read_32(lb++);

        psx_gpu_write_data_mem(dmabuf, size);

        free(dmabuf);
    }
    else if (chcr == DMA_CHAIN)
    {
        psx_gpu_dma_chain((uint32_t*)psx_mem_ram_ptr(0),
                          madr & (RAM_SIZE - 1));
    }
}
