/*
 * Defines the functions used to emulate the R3000 instructions.
 */

#include "address.h"
#include "../bios/bios.h"
#include "../bios/exception.h"
#include "codegate.h"
#include "globals.h"
#include "hardware.h"
#include "interpreter.h"
#include "memory.h"

#include <stdint.h>
#include <sys/time.h>
#include <unistd.h>

/* General purpose registers */
static uint32_t GPR[32] = { 0 };

/* Integer multiplier registers */
static struct
{
    uint32_t hi, lo;
} IMR;

/*
 * Get / set registers.
 */
uint32_t psx_get_reg(int n)
{
    return GPR[n];
}

/* VSync time handling */

#define VSYNC_INTERVAL (1000000 / 60) /* 60 FPS */

static struct timeval vsync_tv = { 0, 0 };

void vsync_time_reset(void)
{
    static struct timezone tz;

    gettimeofday(&vsync_tv, &tz);
}

int vsync_time_elapsed(void)
{
    static struct timezone tz;
    struct timeval tv;

    uint64_t totusec1, totusec2;

    gettimeofday(&tv, &tz);

    totusec1 = tv.tv_sec * 1000000 + tv.tv_usec;
    totusec2 = vsync_tv.tv_sec * 1000000 + vsync_tv.tv_usec;

    return (totusec1 > (totusec2 + VSYNC_INTERVAL));
}

void psx_set_reg(int n, uint32_t v)
{
    GPR[n] = v;
}

/* Execute one instruction */
void psx_exec_one(void)
{
    uint32_t i[4];
    instr_func instr;

    instr = psx_instr_at(pc);
    psx_instr_params_at(pc, i);

    pc += 4;

    instr(i[0], i[1], i[2], i[3]);
}

/* Handle hardware interrupts */
static inline void psx_handle_interrupts(void)
{
    /* if (psx_hardware_read_32(0x1070) & psx_hardware_read_32(0x1074)) */
    if (vsync_time_elapsed())
    {
        psx_bios_exception();
        vsync_time_reset();
    }
}

/* Handles jumps - delay shots, etc. */
static inline void jump(uint32_t addr)
{
    /* TODO : Handle delays, etc. */
    psx_exec_one();

    pc = addr;

    psx_handle_interrupts();
}

static inline void jump_and_link(uint32_t addr)
{
    /* 31 = RA, return address, aka. link register */
    GPR[31] = pc + 4;
    jump(addr);
}

/* Implementation of theinstructions */

#define INSTR(n) \
void psx_instr_##n(uint32_t a, uint32_t b, uint32_t c, uint32_t d)

/* Special instructions : call to BIOS, etc. */
INSTR(callbios)
{
    psx_bios_entry((pc >> 4) - 0xA, GPR);

    /* Jump back to the old PC */
    pc = GPR[RA];
}

/* Real R3000 instructions */

INSTR(add)
{
    if (c)
        GPR[c] = GPR[a] + GPR[b];
}

INSTR(addi)
{
    if (b)
        GPR[b] = GPR[a] + (int16_t)c;
}

INSTR(addiu)
{
    if (b)
        GPR[b] = GPR[a] + (int16_t)c;
}

INSTR(addu)
{
    if (c)
        GPR[c] = GPR[a] + GPR[b];
}

INSTR(and)
{
    if (c)
        GPR[c] = GPR[a] & GPR[b];
}

INSTR(andi)
{
    if (b)
        GPR[b] = GPR[a] & (uint16_t)c;
}

INSTR(beq)
{
    if (GPR[a] == GPR[b])
        jump(adapt_branch_address(c));
}

INSTR(bgez)
{
    if ((int32_t)GPR[a] >= 0)
        jump(adapt_branch_address(c));
}

INSTR(bgezal)
{
    if ((int32_t)GPR[a] >= 0)
        jump_and_link(adapt_branch_address(c));
}

INSTR(bgtz)
{
    if ((int32_t)GPR[a] > 0)
        jump(adapt_branch_address(c));
}

INSTR(blez)
{
    if ((int32_t)GPR[a] <= 0)
        jump(adapt_branch_address(c));
}

INSTR(bltz)
{
    if ((int32_t)GPR[a] < 0)
        jump(adapt_branch_address(c));
}

INSTR(bltzal)
{
    if ((int32_t)GPR[a] < 0)
        jump_and_link(adapt_branch_address(c));
}

INSTR(bne)
{
    if (GPR[a] != GPR[b])
        jump(adapt_branch_address(c));
}

INSTR(break) /* C keyword, but do not care as it is concatenated
                to the prefix */
{
    /* Maybe TODO, PCSX does not seem to handle this instruction... */
}

INSTR(div)
{
    if (GPR[b])
    {
        IMR.lo = (int32_t)a / (int32_t)b;
        IMR.hi = (int32_t)a % (int32_t)b;
    }
}

INSTR(divu)
{
    if (GPR[b])
    {
        IMR.lo = a / b;
        IMR.hi = a % b;
    }
}

INSTR(j)
{
    jump(adapt_j_address(a));
}

INSTR(jal)
{
    jump_and_link(adapt_j_address(a));
}

INSTR(jalr)
{
    /* Save the addr in case we link to RS */
    uint32_t addr = GPR[a];

    if (c)
        GPR[c] = pc;
    jump(addr);
}

INSTR(jr)
{
    jump(GPR[a]);
}

INSTR(lb)
{
    int8_t read = (int8_t)psx_mem_read_8(GPR[a] + (int16_t)c);
    if (b)
        GPR[b] = (int32_t)read;
}

INSTR(lbu)
{
    uint8_t read = psx_mem_read_8(GPR[a] + (int16_t)c);
    if (b)
        GPR[b] = read;
}

INSTR(lh)
{
    int16_t read = (int16_t)psx_mem_read_16(GPR[a] + (int16_t)c);
    if (b)
        GPR[b] = (int32_t)read;
}

INSTR(lhu)
{
    uint16_t read = psx_mem_read_16(GPR[a] + (int16_t)c);
    if (b)
        GPR[b] = read;
}

INSTR(lui)
{
    GPR[b] = ((uint16_t)c) << 16;
}

INSTR(lw)
{
    uint32_t read = psx_mem_read_32(GPR[a] + (int16_t)c);
    if (b)
        GPR[b] = read;
}

INSTR(mfhi)
{
    if (c)
        GPR[c] = IMR.hi;
}

INSTR(mflo)
{
    if (c)
        GPR[c] = IMR.lo;
}

INSTR(mthi)
{
    IMR.hi = GPR[a];
}

INSTR(mtlo)
{
    IMR.lo = GPR[a];
}

INSTR(mult)
{
    uint64_t prod = (int64_t)((int64_t)GPR[a] * (int64_t)GPR[b]);

    IMR.lo = (uint32_t)(prod & 0xFFFFFFFF);
    IMR.hi = (uint32_t)((prod >> 32) & 0xFFFFFFFF);
}

INSTR(multu)
{
    uint64_t prod = (uint64_t)GPR[a] * (uint64_t)GPR[b];

    IMR.lo = (uint32_t)(prod & 0xFFFFFFFF);
    IMR.hi = (uint32_t)((prod >> 32) & 0xFFFFFFFF);
}

INSTR(nor)
{
    if (c)
        GPR[c] = ~(GPR[a] | GPR[b]);
}

INSTR(or)
{
    if (c)
        GPR[c] = GPR[a] | GPR[b];
}

INSTR(ori)
{
    if (b)
        GPR[b] = GPR[a] | c;
}

INSTR(sb)
{
    psx_mem_write_8(GPR[a] + (int16_t)c, GPR[b]);
}

INSTR(sh)
{
    psx_mem_write_16(GPR[a] + (int16_t)c, GPR[b]);
}

INSTR(sll)
{
    if (c)
        GPR[c] = GPR[b] << d;
}

INSTR(sllv)
{
    if (c)
        GPR[c] = GPR[b] << GPR[a];
}

INSTR(slt)
{
    if (c)
        GPR[c] = (int32_t)GPR[a] < (int32_t)GPR[b];
}

INSTR(slti)
{
    if (b)
        GPR[b] = (int32_t)GPR[a] < (int32_t)c;
}

INSTR(sltiu)
{
    if (b)
        GPR[b] = GPR[a] < c;
}

INSTR(sltu)
{
    if (c)
        GPR[c] = GPR[a] < GPR[b];
}

INSTR(sra)
{
    if (c)
        GPR[c] = (int32_t)GPR[b] >> d;
}

INSTR(srav)
{
    if (c)
        GPR[c] = (int32_t)GPR[b] >> GPR[a];
}

INSTR(srl)
{
    if (c)
        GPR[c] = GPR[b] >> d;
}

INSTR(srlv)
{
    if (c)
        GPR[c] = GPR[b] >> GPR[a];
}

INSTR(sub)
{
    if (c)
        GPR[c] = GPR[a] - GPR[b];
}

INSTR(subu)
{
    if (c)
        GPR[c] = GPR[a] - GPR[b];
}

INSTR(sw)
{
    psx_mem_write_32(GPR[a] + (int16_t)c, GPR[b]);
}

INSTR(syscall)
{
    /* TODO */
}

INSTR(xor)
{
    if (c)
        GPR[c] = GPR[a] ^ GPR[b];
}

INSTR(xori)
{
    if (b)
        GPR[b] = GPR[a] ^ c;
}
