/*
 * Functions handling read/write access to the PSX memory.
 */

#include "globals.h"
#include "hardware.h"
#include "memory.h"

#include <stdint.h>
#include <stdlib.h>
#include <string.h>

/*
 * Inspired from how PCSX handles memory references.
 *
 * This map adressed by a 16-bit integer contains the base address of the
 * memory zones. For example, the adress 0x42421337 can be accessed via :
 *   RMEMREF[0x4242][0x1337]
 * 
 * This allow fast decoding of memory adresses.
 */
static uint8_t** RMEMREF = NULL;
static uint8_t** WMEMREF = NULL;

/* RAM - 2M */
static uint8_t* RAM = NULL;

/* Parallel port - 64K */
static uint8_t* PPORT = NULL;

/* Hardware - 12K */
static uint8_t* HW = NULL;

/* BIOS - 512K */
static uint8_t* BIOS = NULL;

void psx_mem_init(void)
{
    int i;

    RAM = malloc(RAM_SIZE);
    PPORT = malloc(PPORT_SIZE);
    HW = malloc(HW_SIZE);
    BIOS = malloc(BIOS_SIZE);

    /* Copy the code segment to the RAM */
    memcpy(RAM + (t_seg_addr & (RAM_SIZE - 1)), t_seg, t_seg_size);

    RMEMREF = malloc(0x10000 * sizeof (uint8_t*));
    WMEMREF = malloc(0x10000 * sizeof (uint8_t*));

    /* Fill the MEMREFs */
    for (i = 0; i < 256; ++i)
        RMEMREF[i] = &RAM[(i & 0x1F) << 16];

    memcpy(RMEMREF + 0x8000, RMEMREF, 256 * sizeof (uint8_t*));
    memcpy(RMEMREF + 0xA000, RMEMREF, 256 * sizeof (uint8_t*));

    for (i = 0; i < 1; ++i)
    {
        RMEMREF[0x1F00 + i] = &PPORT[i << 16];
        RMEMREF[0x1F80 + i] = &HW[i << 16];
    }

    for (i = 0; i < 8; ++i)
        RMEMREF[0xBFC0 + i] = &BIOS[i << 16];

    memcpy(WMEMREF, RMEMREF, 0x10000 * sizeof (uint8_t*));
}

void psx_mem_shutdown(void)
{
    free(RAM);
    free(PPORT);
    free(HW);
    free(BIOS);
    free(RMEMREF);
    free(WMEMREF);
}

uint8_t psx_mem_read_8_direct(uint32_t a)
{
    return RMEMREF[a >> 16][a & 0xFFFF];
}

uint16_t psx_mem_read_16_direct(uint32_t a)
{
    return *((uint16_t*)&RMEMREF[a >> 16][a & 0xFFFF]);
}

uint32_t psx_mem_read_32_direct(uint32_t a)
{
    return *((uint32_t*)&RMEMREF[a >> 16][a & 0xFFFF]);
}

void psx_mem_write_8_direct(uint32_t a, uint8_t d)
{
    WMEMREF[a >> 16][a & 0xFFFF] = d;
}

void psx_mem_write_16_direct(uint32_t a, uint16_t d)
{
    *((uint16_t*)&WMEMREF[a >> 16][a & 0xFFFF]) = d;
}

void psx_mem_write_32_direct(uint32_t a, uint32_t d)
{
    *((uint32_t*)&WMEMREF[a >> 16][a & 0xFFFF]) = d;
}

uint8_t psx_mem_read_8(uint32_t addr)
{
    uint16_t u, l;

    u = addr >> 16;
    l = addr & 0xFFFF;

    if (u == 0x1F80)
        return psx_hardware_read_8(l);
    else
        return RMEMREF[u][l];
}

uint16_t psx_mem_read_16(uint32_t addr)
{
    uint16_t u, l;

    u = addr >> 16;
    l = addr & 0xFFFF;

    if (u == 0x1F80)
        return psx_hardware_read_16(l);
    else
        return *((uint16_t*)&RMEMREF[u][l]);
}

uint32_t psx_mem_read_32(uint32_t addr)
{
    uint16_t u, l;

    u = addr >> 16;
    l = addr & 0xFFFF;

    if (u == 0x1F80)
        return psx_hardware_read_32(l);
    else
        return *((uint32_t*)&RMEMREF[u][l]);
}

void psx_mem_write_8(uint32_t addr, uint8_t data)
{
    uint16_t u, l;

    u = addr >> 16;
    l = addr & 0xFFFF;

    if (u == 0x1F80)
        psx_hardware_write_8(l, data);
    else
        WMEMREF[u][l] = data;
}

void psx_mem_write_16(uint32_t addr, uint16_t data)
{
    uint16_t u, l;

    u = addr >> 16;
    l = addr & 0xFFFF;

    if (u == 0x1F80)
        psx_hardware_write_16(l, data);
    else
        *((uint16_t*)&WMEMREF[u][l]) = data;
}

void psx_mem_write_32(uint32_t addr, uint32_t data)
{
    uint16_t u, l;

    u = addr >> 16;
    l = addr & 0xFFFF;

    if (u == 0x1F80)
        psx_hardware_write_32(l, data);
    else
        *((uint32_t*)&WMEMREF[u][l]) = data;
}

uint8_t* psx_mem_ram_ptr(uint32_t addr)
{
    return &RAM[addr];
}
