(*
 * Module to create the code graph of a Psx_file.t.
 *)
module D = R3000_decoder
module P = Psx_file

type node =
|   CodeBlock of D.instr list * t
|   Branching of D.instr list * t list
and t = node ref

type intern_node =
|   CodeBlock_ of D.instr list * int
|   Branching_ of D.instr list * int list
and intern_t = intern_node ref

let addr_to_instr_off f a =
    let lb = f.P.header.P.text.P.addr in
    Int32.to_int (Int32.shift_right_logical (Int32.sub a lb) 2)

let is_br_instr i = match D.decode i with
|   D.J _
|   D.Beq _
|   D.Bgez _
|   D.Bgezal _
|   D.Bgtz _
|   D.Blez _
|   D.Bltz _
|   D.Bltzal _
|   D.Bne _
|   D.Jal _
|   D.Jalr _
|   D.Jr _ ->
        true
|   _ ->
        false

let rec find_next_non_det c off =
    if off >= (Array.length c) then
        off - 1
    else if is_br_instr c.(off) then
        off - 1
    else
        find_next_non_det c (off + 1)

let force_opt = function
|   Some x -> x
|   None -> assert false

(*
 * Find the first offset in the map with the same node.
 *)
let rec find_first map off =
    if off = 0 then
        0
    else if map.(off - 1) <> map.(off) then
        off
    else
        find_first map (off - 1)

(*
 * Splits a node into two nodes.
 *)
let split map c off =
    let ol, tl = Array.of_list, Array.to_list in
    let instrs = function
    |   Some { contents = CodeBlock_ (code, li) } -> ol code, li
    |   _ -> assert false
    in
    if (off = 0) || (map.(off - 1) <> map.(off)) then
        force_opt map.(off)
    else
        let first = find_first map off in
        let prevoff = off - first in
        let previnstrs, li = instrs map.(first) in
        let instrs_b = Array.init prevoff (fun i -> previnstrs.(i)) in
        let instrs_a = Array.init
            ((Array.length previnstrs) - prevoff)
            (fun i -> previnstrs.(prevoff + i - 1))
        in
        let newnode = ref (CodeBlock_ (tl instrs_a, li)) in
        let () =
            for i = first to off - 1 do
                (force_opt map.(i)) := CodeBlock_ (tl instrs_b, off)
            done
        in
        let () =
            for i = off to (off + (Array.length instrs_a) - 1) do
                map.(i) <- Some newnode
            done
        in
        newnode

let get_two c off =
    try
        [D.decode c.(off) ; D.decode c.(off + 1)]
    with Invalid_argument _ ->
        [D.decode c.(off)]

let rec make_from map f c off =
    if map.(off) <> None then
        split map c off
    else if is_br_instr c.(off) then
        make_br map f c off
    else
        let endoff = find_next_non_det c off in
        let ninstr = endoff - off + 1 in
        let code = Array.init ninstr (fun n -> D.decode c.(off + n)) in
        let node = ref (CodeBlock_ (Array.to_list code, endoff + 1)) in
        let () =
            for i = off to endoff do
                map.(i) <- Some node
            done
        in
        let _ = make_from map f c (endoff + 1) in
        node

and make_br map f c off = match (D.decode c.(off)) with
    (*
     * Here is the real part of the job.
     *
     * Currently, only handles positive B* branchings and uncond jumps.
     *)
|   D.Beq i
|   D.Bne i ->
        make_generic_br map f c off (i.D.I.imm)
|   D.Bgez b
|   D.Bgezal b
|   D.Bgtz b
|   D.Blez b
|   D.Bltz b
|   D.Bltzal b ->
        make_generic_br map f c off (b.D.B.imm)
|   D.J j
|   D.Jal j ->
        let addr = Int32.of_int j.D.J.target in
        let base = Int32.logand 0xF0000000l (f.P.header.P.text.P.addr) in
        let shifted = Int32.mul addr 4l in
        let final_addr = Int32.add base shifted in
        let next_off = addr_to_instr_off f final_addr in
        let node = ref (Branching_ (get_two c off, [next_off])) in
        map.(off) <- Some node ;
        let _ = make_from map f c next_off in
        if (D.decode c.(off)) = (D.Jal j) then (
            (*
             * We can safely assume that Jal will be followed by a Jr ra.
             *)
            node := Branching_ (get_two c off, [next_off ; off + 2]) ;
            ignore (make_from map f c (off + 2))
        ) ;
        node
|   _ ->
        let n = ref (Branching_ (get_two c off, [])) in
        map.(off) <- Some n ;
        n

and make_generic_br map f c off next_off =
    let next_off =
        if next_off > 32767 then
            next_off - 65535
        else
            next_off
    in
    let next = off + next_off in
    let node = ref (Branching_ (get_two c off, [off + 2 ; next])) in
    map.(off) <- Some node ;
    let _ = make_from map f c (off + 2) in
    let _ = make_from map f c next in
    node

(*
 * Conversion from intern_t to t.
 *)

let rec build_nodemap map nodemap off = match map.(off), nodemap.(off) with
|   None, _ ->
        assert false
|   _, Some n ->
        ()
|   Some cg, None -> (
        match !cg with
        |   CodeBlock_ (instrs, next) ->
                let x = ref (CodeBlock (instrs, Obj.magic ())) in
                nodemap.(off) <- Some x ;
                build_nodemap map nodemap next
        |   Branching_ (instr, li) ->
                let x = ref (Branching (instr, [])) in
                nodemap.(off) <- Some x ;
                List.iter (build_nodemap map nodemap) li
    )

(* Small utilities *)
let cb_next = function
|   CodeBlock_ (_, next) -> next
|   _ -> assert false

let br_next = function
|   Branching_ (_, li) -> li
|   _ -> assert false

let cb_with_next n = function
|   CodeBlock (i, _) -> CodeBlock (i, n)
|   _ -> assert false

let br_with_next l = function
|   Branching (i, _) -> Branching (i, l)
|   _ -> assert false

let rec build_graph map nodemap off =
    let oldnode = force_opt map.(off) in
    let node = force_opt nodemap.(off) in
    match !node with
    |   CodeBlock (instrs, x) when x = (Obj.magic ()) ->
            let noff = cb_next !oldnode in
            node := cb_with_next (force_opt nodemap.(noff)) !node ;
            ignore (build_graph map nodemap noff) ;
            node
    |   Branching (instrs, x) when x = [] ->
            let noff = br_next !oldnode in
            let n = List.map (fun o -> force_opt nodemap.(o)) noff in
            node := br_with_next n !node ;
            List.iter (fun o -> ignore (build_graph map nodemap o)) noff ;
            node
    |   _ -> node

let intern_to_extern map off =
    let nodemap = Array.make (Array.length map) None in
    ignore (build_nodemap map nodemap off) ;
    build_graph map nodemap off

let make f =
    (* Allocates a map of the text segment *)
    let map = Array.make (Array.length f.P.code) None in

    (* Start address = init_pc *)
    let pc = addr_to_instr_off f f.P.header.P.init_pc in
    let _ = make_from map f f.P.code pc in
    intern_to_extern map pc
