(*
 * Compiles a Psx_file.t to unoptimized LLVM bytecode.
 *
 * Basically, it translates every instruction to a function call.
 *)

#include "../config.h"

module D = R3000_decoder
module P = Psx_file
module Reg = R3000_reg

type state = {
    ctx : Llvm.llcontext ;
    m : Llvm.llmodule ;
    b : Llvm.llbuilder ;
    mutable pc : Llvm.llvalue ;
    f : Psx_file.t
}

(* LLVM is imperative, we'll use the module in the state to output code *)
type out = state

let init f =
    let ctx = Llvm.create_context () in
    let b = Llvm.builder ctx in
    let m = Llvm.create_module ctx "mmorph_main" in
    { ctx = ctx ; m = m ; f = f ; b = b ; pc = Obj.magic () }

let to_bytecode o chan =
    Llvmext.write_bitcode_to_chan o.m chan

let copy ic oc =
    let eof = ref false in
    let buf = String.create 1024 in
    while not !eof do
        let l = input ic buf 0 1024 in
        if l <> 1024 then
            eof := true ;
        output oc buf 0 l
    done

let to_binary o chan =
    let llc_out, llc_in = Unix.open_process (LLC_PATH ^ " -filetype=asm") in
    let out = Filename.temp_file "translator" ".exe" in
    let gcc_cmd = Printf.sprintf "%s -x %s -o %s -L%s -lpsx -lbios -"
        CC_PATH "assembler" out LIBDIR in
    let gcc_out, gcc_in = Unix.open_process gcc_cmd in
    to_bytecode o llc_in ;
    close_out llc_in ;
    copy llc_out gcc_in ;
    ignore (Unix.close_process (gcc_out, gcc_in)) ;
    copy (open_in_bin out) chan ;
    Unix.unlink out

let to_il o chan =
    Llvmext.print_module o.m chan

(* ---------------------------------------------------------------------- *)

module I = struct
    let to_llvm_args a =
        let rs = Reg.to_int a.D.I.rs in
        let rt = Reg.to_int a.D.I.rt in
        (rs, rt, a.D.I.imm, 0)
end

module J = struct
    let to_llvm_args a = (a.D.J.target, 0, 0, 0)
end

module R = struct
    let to_llvm_args a =
        let rs = Reg.to_int a.D.R.rs in
        let rt = Reg.to_int a.D.R.rt in
        let rd = Reg.to_int a.D.R.rd in
        (rs, rt, rd, a.D.R.shift)
end

module S = struct
    let to_llvm_args a = (a.D.S.code, 0, 0, 0)
end

module B = struct
    let to_llvm_args a =
        let rs = Reg.to_int a.D.B.rs in
        (rs, a.D.B.imm, 0, 0)
end

let add_instr_functions st =
    let vt = Llvm.void_type st.ctx in
    let it = Llvm.i32_type st.ctx in
#define R3000_INSTR(name, args) \
    ignore ( \
        Llvm.declare_function ("psx_instr_" ^ (String.lowercase #name)) \
            (Llvm.function_type vt (Array.make 4 it)) \
            st.m \
    ) ;
#include "../disasm/instrs.inl"
#undef R3000_INSTR
    ()

let define_pc st =
    let it = Llvm.i32_type st.ctx in
    let pc = Int64.of_int32 st.f.P.header.P.init_pc in
    let v = Llvm.const_of_int64 it pc false in
    ignore (Llvm.define_global "pc" v st.m)

let define_segs st =
    let it = Llvm.i32_type st.ctx in
    let bt = Llvm.i8_type st.ctx in
    let def_i32 n v =
        let v = Int64.of_int32 v in
        ignore (Llvm.define_global n (Llvm.const_of_int64 it v false) st.m)
    in
    let def_seg n v =
        let l = Array.length v in
        let arr = Array.init (4 * l) (fun i ->
            let shift = [| 0 ; 8 ; 16 ; 24 |] in
            let x = Int32.shift_right_logical v.(i / 4) shift.(i mod 4) in
            let x = Int32.logand x 0xFFl in
            Llvm.const_int bt (Int32.to_int x)
        ) in
        let v = Llvm.const_array bt arr in
        ignore (Llvm.define_global n v st.m)
    in
    def_seg "t_seg" st.f.P.code ;
    def_i32 "t_seg_addr" st.f.P.header.P.text.P.addr ;
    def_i32 "t_seg_size" st.f.P.header.P.text.P.size ;
    def_i32 "s_seg_addr" st.f.P.header.P.stack.P.addr ;
    def_i32 "init_gp" st.f.P.header.P.init_gp

external force_some : 'a option -> 'a = "%field0"

let instr_to_llvm_fun st i =
    let n = match i with
#define R3000_INSTR(name, args) \
    |   D.name _ -> "psx_instr_" ^ (String.lowercase #name)
#include "../disasm/instrs.inl"
#undef R3000_INSTR
    in
    force_some (Llvm.lookup_function n st.m)

let make_table st =
    let c = st.f.P.code in
    let it = Llvm.i32_type st.ctx in
    let l = Array.length c in
    let d = Array.map (fun i ->
        try
            D.decode i
        with Invalid_argument _ ->
            D.decode 0l
    ) c in
    let instrs = Array.map (instr_to_llvm_fun st) d in
    let params = Array.init (4 * l) (fun i ->
        let a, b, c, d = match d.(i / 4) with
#define R3000_INSTR(name, args) \
        |   D.name a -> args.to_llvm_args a
#include "../disasm/instrs.inl"
#undef R3000_INSTR
        in
        let v = [| a ; b ; c ; d |].(i mod 4) in
        Llvm.const_int it v
    ) in
    instrs, params

let generate_instrs_and_params st =
    let instrs, params = make_table st in
    let iv = Llvm.const_array (Llvm.type_of instrs.(0)) instrs in
    let pv = Llvm.const_array (Llvm.type_of params.(0)) params in
    ignore (Llvm.define_global "instrs" iv st.m) ;
    ignore (Llvm.define_global "instrs_params" pv st.m)

let compile st =
    add_instr_functions st ;
    define_pc st ;
    define_segs st ;
    generate_instrs_and_params st ;
    st
