module ComputerPsxFile = Psx_file.Make (Os_fs)

type cfg = {
    mutable fname : string ;
    mutable act : Psx_file.t -> out_channel -> unit ;
    mutable out : string
}

let uniqid cnt =
    let id = "n" ^ (string_of_int !cnt) in
    incr cnt ;
    id

let ilist_to_s code =
    let cr = List.map R3000_decoder.to_string code in
    List.fold_left (fun a b -> a ^ "\\n" ^ b)
        (List.hd cr) (List.tl cr)

let fpf = Printf.fprintf
let rec dump_dot p cnt h cg =
    if Hashtbl.mem h cg then
        Hashtbl.find h cg
    else match !cg with
    |   Codegraph.CodeBlock (code, next) ->
            let id = uniqid cnt in
            fpf p "%s [label=\"%s\"]\n" id (ilist_to_s code) ;
            Hashtbl.add h cg id ;
            let nid = dump_dot p cnt h next in
            fpf p "%s -> %s\n" id nid ;
            id
    |   Codegraph.Branching (instrs, next) ->
            let id = uniqid cnt in
            fpf p "%s [label=\"%s\",shape=box]\n" id
                (ilist_to_s instrs) ;
            Hashtbl.add h cg id ;
            List.iter (fun n ->
                let nid = dump_dot p cnt h n in
                Hashtbl.add h n nid ;
                fpf p "%s -> %s\n" id nid
            ) next ;
            id

let dump_graph_dot f out =
    let cg = Codegraph.make f in
    let cnt = ref 0 in
    let s = Hashtbl.create 2048 in
    fpf out "digraph G {\n" ;
    ignore (dump_dot out cnt s cg) ;
    fpf out "}\n"

let dump_backend_out typ f out =
    let module Backend = Llvmunoptbackend in
    let outfun = match typ with
    |   `Binary -> Backend.to_binary
    |   `Bytecode -> Backend.to_bytecode
    |   `Il -> Backend.to_il
    in
    let st = Backend.init f in
    let o = Backend.compile st in
    outfun o out

let parse_args () =
    let usage = "usage: translator [-dot|-bc|-il] <filename>" in
    let n = Obj.magic () in
    let cfg = { fname = n ; act = dump_backend_out `Binary ; out = n } in
    let opts = [
        "-dot", Arg.Unit (fun () -> cfg.act <- dump_graph_dot),
            "dump the code graph in dot format" ;
        "-bc", Arg.Unit (fun () -> cfg.act <- dump_backend_out `Bytecode),
            "dump the LLVM bytecode" ;
        "-il", Arg.Unit (fun () -> cfg.act <- dump_backend_out `Il),
            "dump the LLVM intermediate language" ;
        "-o", Arg.String (fun s -> cfg.out <- s),
            "path to the output file (default: ./a.out)"
    ] in
    Arg.parse opts (fun s -> cfg.fname <- s) usage ;
    if cfg.fname == n then (
        Arg.usage opts usage ;
        exit 1
    ) ;
    if cfg.out == n then (
        cfg.out <- "./a.out"
    ) ;
    cfg

let main cfg =
    let stream = Os_fs.open_ cfg.fname in
    let psxfile = ComputerPsxFile.parse stream in
    Os_fs.close stream ;
    let out = open_out cfg.out in
    cfg.act psxfile out ;
    close_out out

let () =
    let cfg = parse_args () in
    main cfg
